import numpy as np
import bren.autodiff.nodes as nodes
import bren as br


class Graph(object):

	"""
	The Graph object helps keep track of all of the operations performed on a Variable object. In doing so, it performs a backward pass and computes the gradients of all of the Variables involved in the computation graph. 
	"""

	_g = None

	def __init__(self) -> None:
		self.nodes = set()
		Graph._g = self.nodes

	def __reset_counts(self, root):
		if hasattr(root, "count"):
			root.count = 0
		else: 	
			for child in root.__subclasses__():
				self.__reset_counts(child)

	def __reset_session(self):
		try:
			del Graph._g
		except: pass

		Graph._g = None
		self.__reset_counts(nodes.Node)
		self.__reset_counts(nodes.Operator)

	def __enter__(self):
		Graph._g.add("entered graph")
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		self.__reset_session()

	def __g(self, dy, dx):
		op = [dy._source]
		dy._source.gradient = np.ones(shape=dy.shape, dtype=dy.dtype)

		for x in dx:
			x._source.gradient = 0 

		for operation in op:
			try:
				grads = operation(dout=operation.gradient)

				for inp, grad in zip(operation.inputs, grads):
					if inp: inp.gradient += grad

					if isinstance(inp, nodes.Operator):
						op.append(inp)

			except AttributeError:
				raise Exception("Cannot find gradient of 2 unconnected variables")

		for value in dx:
			yield br.Constant(value._source.gradient, dtype=value.dtype)

	def grad(self, dy, dx): 
		"""
		Performs a backward pass and returns a list of 
		"""
		return list(self.__g(dy, dx))



from bren.nn.layers import Layer
import numpy as np


class Flatten(Layer):
    """
    The `Flatten` Layer flattens (reduces the number of the dimensions of the array to 1) the input Array.

    Parameters
    ----------
    name (`str`): The name of the layer.
    """

    def __init__(self, name=None, **kwargs) -> None:
        super().__init__(name, **kwargs)

    def call(self, x):
        return x.flatten()[..., np.newaxis]
    




"""
bren
====
bren is a custom [numpy](https://numpy.org) based library, powered by automatic differentiation,
inspired by [Tensorflow](https://www.tensorflow.org)/[Keras](https://keras.io), which allows users to build small scale simple neural networks. It's analogous yet simpler design to the Keras api allows users to produce, train and save their own models, with custom components, without having to learn an entirely new structure. 

bren is part of a sequence of neural network from scratch projects and a successor to the [neural-network-from-scratch-v2](https://github.com/OmPanchal/nn-from-scratch-2), with one major update being the integration of automatic differentitation. Automatic differentiation allows for the real-time determination of derivatives during back propagation and reduces the need for users to couple mathematical computation with derivatives as was required in the previous projects.

Examples tend to import `bren` as `br`
```python
import bren as br

A = br.Variable([1, 2, 3])
print(A + 2) # <Variable value=[3. 4. 5.] dtype=float64>
```

Autodiff
--------
Bren is an autodiff driven neural network library, with backpropagation making use of `br.Graph` to find the derivatives of the trainable parameters with respect to the loss. This is governed by `br.Variable` which keeps track of any operation which have been performed on the `Variable` object.

```python
import bren as br

A = br.Variable([1, 2, 3], dtype="float64")

with br.Graph() as g:
	B = A ** 2 # any computation performed on a Variable will be tracked in a with statement
	print(g.grad(B, [A])) # [<Constant value=[2. 4. 6.] dtype=float64>]
```
"""


from bren.core.core import Constant, Variable
from bren.autodiff.nodes.Graph import Graph
import bren.nn as nn
from bren.autodiff.operations.ops import custom_gradient



__all__ = [nn, Constant, custom_gradient, Graph, Variable]


import numpy as np
from bren.nn.initialisers.Initialiser import Initialiser


class HeNormal(Initialiser):
	"""
	Returns values taken from a normal distribution with mean 0 and a stddev of `sqrt(2 / (fan_in))` where `fan_in` is a the number of inputs 
	"""

	def __init__(self, shape, dtype="float32", **kwargs) -> None:
		super().__init__(shape, dtype, **kwargs)

		self.mean = kwargs.get("mean") or 0
		self.fan_in, _ = kwargs.get("inout") or shape
		self.variance = 2 / self.fan_in

	def call(self, *args, **kwargs):
		return np.random.normal(self.mean, self.variance, size=self.shape)


from bren.nn.initialisers import Initialiser
import numpy as np


class GlorotNormal(Initialiser):
	"""
	Returns values taken from a normal distribution with mean 0 and a stddev of `sqrt(2 / (fan_in + fan_out))` where `fan_in` is a the number of inputs while `fan_out` is the number of outputs
	"""

	def __init__(self, shape, dtype="float32", **kwargs) -> None:
		super().__init__(shape, dtype, **kwargs)

		self.mean = kwargs.get("mean") or 0
		self.fan_in, self.fan_out = kwargs.get("inout") or shape
		fan_avg = (self.fan_in + self.fan_out) / 2
		self.stddev = np.sqrt(1 / fan_avg)

	def call(self, *args, **kwargs):
		return np.random.normal(self.mean, self.stddev, size=self.shape)



from bren.nn.layers.Layer import Layer


class Linear(Layer):
	"""
	The linear activation function perform no operation on the input

	Parameters
	----------
	x (`br.Variable`): The input Array
	name (`str`): The name of the activation.
	"""

	def __init__(self, name=None, **kwargs) -> None:
		super().__init__(name, **kwargs)

	def call(self, x): return x



from bren import Variable
import numpy as np


class Metric(object):
	"""
	The base `Metric` class
	"""

	def __init__(self) -> None: 
		self.__reset = np.vectorize(self.__reset, cache=False)
		self.variables = []

	def __reset(self, var): var.assign(0) 

	def update(self, y_pred, y_true, weights=None):
		"""
		The `update` function is called at the end of every pass through of training data.

		Parameters
		----------
		y_pred (br.Variable): the predicted values from the model
		y_true (br.Variable): the actual values
		"""
		...

	def reset(self):
		"""
		Resets all of the local Variables.
		"""
		self.__reset(self.variables)

	def add_weight(self, val, **kwargs):
		"""
		Creates a new Variable.

		Parameter
		---------
		val (`list`, `np.ndarray`): The value of the Variable. 
		"""
		var = Variable(val, **kwargs)
		self.variables.append(var)
		return var
	
	def result(self):
		"""
		Returns the final current value of the metrics
		"""
		...

	def __call__(self, y_pred, y_true, **kwargs): 
		self.update(y_pred, y_true)
		return self.result(**kwargs)




from bren.nn.optimisers import Optimiser
from bren import Constant
import numpy as np


class Adam(Optimiser):
	def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999, epsilon=10e-10) -> None:
		super().__init__()
		self.learning_rate = learning_rate
		self.beta1 = beta1
		self.beta2 = beta2
		self.epsilon = epsilon
		self.s = []
		self.m = []

	def initialise(self, len):
		for _ in range(len):
			self.s.append(Constant(0))
			self.m.append(Constant(0))
		 
	def apply_gradients(self, vars, grads, **kwargs):
		if len(self.s) == 0 and len(self.m) == 0:
			self.initialise(len(vars))

		self.update(vars, grads, self.m, self.s)

	def update(self, var, grad, m, s):
		_m = m.assign((self.beta1 * m) - ((1 - self.beta1) * grad))
		_s = s.assign((self.beta2 * s) + (1 - self.beta2) * (grad ** 2))

		m_hat = _m / (1 - self.beta1)
		s_hat = _s / (1 - self.beta2) 

		_t = (self.learning_rate * m_hat) / np.sqrt(s_hat + self.epsilon)

		var.assign_add(_t)
	


from bren.nn.layers import Layer
import numpy as np
from bren.autodiff.operations.ops import custom_gradient


class Softmax(Layer):
    """
    Returns a Array of values of the same size as its inputs where `tmp = np.exp(x)` and `tmp / np.sum(tmp)` is returned.

    Parameters
    ----------
    name (`str`): The name of the Layer.
    """

    def __init__(self, name=None, **kwargs) -> None:
        super().__init__(name, **kwargs)

    def call(self, x):
        tmp = np.exp(x)
        return tmp / np.sum(tmp)


class Node(object):
	def __init__(self, value, name) -> None:
		self.value = value
		self._gradient = 0
		self.name = name

	def set_name(self, cls, name=None): return f"{name or cls.__name__}/{cls.count}"

	@property
	def gradient(self): return self._gradient
	@gradient.setter
	def gradient(self, value):
		self._gradient = value


import numpy as np
from bren.autodiff.nodes import (Operator, Graph)
import inspect


np.seterr(invalid="ignore", divide="ignore")

OPS = {}

def custom(grad, funcname=None):
	class operator(Operator):
		count = 0

		def __init__(self, inputs, value=None, name=funcname, **kwargs) -> None:
			super().__init__(inputs, value, name, **kwargs)
			if Graph._g: Graph._g.add(self)
			self.name = self.set_name(operator, name=name)
			self.grad = grad
			operator.count += 1

	return operator


def custom_gradient(grad):
	"""
	A decorator which maps a custom gradient to a function.
	[NOTE] This is advised for functions which perform logical operations.
	
	Parameters 
	----------
	grad: The gradient of the function which is being decorated.

	"""
	def decorator(func):
		OPS[func.__name__ + " (vectorized)"] = custom(grad)
		return np.frompyfunc(func, nin=inspect.signature(func).parameters.__len__(), nout=1)

	return decorator 


def ufunc_grad(*_ufunc_name):
	def decorator(grad):
		for name in _ufunc_name:
			OPS[name] = custom(grad, funcname=name)
		return grad

	return decorator


@ufunc_grad("nongrad")
def default_grad(*ab, dout, **kwargs): return dout, dout

@ufunc_grad(np.abs.__name__, np.absolute.__name__)
def absolute_grad(a, dout, **kwargs): return [np.multiply(dout, (a / np.abs(a)))]

@ufunc_grad(np.add.__name__)
def add_grad(a, b, dout, **kwargs): return dout, dout

@ufunc_grad(np.arccos.__name__)
def arccos_grad(a, dout, **kwargs): return [np.multiply(dout, -(1 / (np.sqrt(1 - (a ** 2)))))]
@ufunc_grad(np.arccosh.__name__)
def arccosh_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / (np.sqrt((a ** 2) - 1))))]

@ufunc_grad(np.arcsin.__name__)
def arcsin_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / (np.sqrt(1 - (a ** 2)))))]
@ufunc_grad(np.arcsinh.__name__)
def arcsinh_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / (np.sqrt((a ** 2) + 1))))]

@ufunc_grad(np.arctan.__name__)
def arctan_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / ((a ** 2) + 1)))]
@ufunc_grad(np.arctanh.__name__)
def arctanh_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / (1 - (a ** 2))))]

@ufunc_grad(np.cbrt.__name__)
def cbrt_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / ((3 * a) ** (2 / 3))))]

@ufunc_grad(np.cos.__name__)
def cos_grad(a, dout, **kwargs): return [np.multiply(dout, -np.sin(a))]
@ufunc_grad(np.cosh.__name__)
def cosh_grad(a, dout, **kwargs): return [np.multiply(dout, np.sinh(a))]

@ufunc_grad(np.divide.__name__)
def divide_grad(a, b, dout, **kwargs):
	return (np.divide(dout, b), -np.multiply(dout, (a / (b ** 2))))

@ufunc_grad(np.exp2.__name__)
def exp2_grad(a, dout, **kwargs): return [np.multiply(dout, (np.log(2) * (2 ** a)))]

@ufunc_grad(np.exp.__name__)
def exp_grad(a, dout, **kwargs): return [np.multiply(dout, np.exp(a))]

@ufunc_grad(np.log10.__name__)
def log10_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / (np.log(10) * a)))]

@ufunc_grad(np.log1p.__name__)
def log1p_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / (a + 1)))]

@ufunc_grad(np.log2.__name__)
def log2_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / np.log(2) * a))]

@ufunc_grad(np.log.__name__)
def log_grad(a, dout, **kwargs): return [np.multiply(dout, (1 / a))]

@ufunc_grad(np.matmul.__name__)
def matmul_grad(a, b, dout, **kwargs):
	return np.matmul(dout, b.T), np.matmul(a.T, dout)

@ufunc_grad(np.multiply.__name__)
def multiply_grad(a, b, dout, **kwargs): return np.multiply(dout, b), np.multiply(dout, a)

@ufunc_grad(np.negative.__name__)
def negative_grad(a, dout, **kwargs): return [np.negative(dout)]

@ufunc_grad(np.power.__name__)
def power_grad(a, b, dout, **kwargs):
	return (np.multiply(dout, (b * (a ** (b - 1)))),
	np.nan_to_num(np.multiply(dout, (a ** b) * (np.log(a))), nan=0, neginf=0, posinf=0))

@ufunc_grad(np.reciprocal.__name__)
def reciprocal_grad(a, dout, **kwargs): return [np.multiply(dout, -np.reciprocal(a ** 2))]

@ufunc_grad(np.sin.__name__)
def sin_grad(a, dout, **kwargs): return [np.multiply(dout, np.cos(a))]
@ufunc_grad(np.sinh.__name__)
def sinh_grad(a, dout, **kwargs): return [np.multiply(dout, np.cosh(a))]

@ufunc_grad(np.sqrt.__name__)
def sqrt(a, dout, **kwargs): return [np.multiply(dout, 1 / (2 * np.sqrt(a)))]

@ufunc_grad(np.subtract.__name__)
def subtract_grad(a, b, dout, **kwargs): return dout, np.negative(dout)

@ufunc_grad(np.tan.__name__)
def tan_grad(a, dout, **kwargs): 
	return [np.multiply(dout, 1 + (np.tan(a) ** 2))]

@ufunc_grad(np.tanh.__name__)
def tanh_grad(a, dout, **kwargs): return [np.multiply(dout, 1 - (np.tanh(a) ** 2))]

@ufunc_grad(np.transpose.__name__)
def transpose_grad(a, dout, **kwargs): return [dout.T]

@ufunc_grad(np.mean.__name__)
def mean_grad(a, dout, **kwargs): return [np.divide(dout, a.size)]



from bren.nn.layers import Layer
from bren.nn.initialisers import get_initialiser, Initialiser, initialiser_from_func
from bren.nn.activations import get_activation, Activation
from bren.nn.utils import rename_key


def set_activation(activation, custom_obs={}):
	out = None

	if type(activation) == str or activation is None: 
		try: out = get_activation(activation)()
		except KeyError:
			try: out = set_activation(custom_obs[activation])
			except KeyError:
				raise KeyError(f"custom object {activation} cannot be found") 
			
	elif issubclass(type(activation), Layer):
		out = activation
	elif type(activation).__name__ == "function":
		out = Activation(activation)
	else: raise AttributeError(f"{activation} is of invalid type for activation")

	return out

def set_initialiser(initialiser):
	out = None

	if type(initialiser) == str:
		out = get_initialiser(initialiser)
	elif issubclass(type(initialiser), Initialiser):
		out = initialiser 
	elif type(initialiser).__name__ == "function":
		out = initialiser_from_func(initialiser)
	else: raise AttributeError(f"{initialiser} is of invalid type for initialisers")

	return out
	

class FullyConnected(Layer):

	"""
	The `FullyConnected` layer returns a layer with a specified amount of `units` where each unit is conceptually connected to every other unit in the previous layer.

	Parameters
	----------
	units (`int`): The number of units for the layer.
	activation (`str`, `Activation`, `function`): the activation function for the layer. After the weights and biases have been applied to the inputs, the result will then be passed through the specified activation function. By default, this is set to `None` which is equivalent to the `Linear` activation function.
	weights	 
	weights_initialiser (`str`, `Initialiser`, `function`): The initialiser function of the weights.
	bias_initialiser (`str`, `Initialiser`, `function`): The initialiser function of the biases.
	use_bias (`bool`): Determines the inclusion of the bias.
	name ("str"): the name of the `FullyConnected` Layer.
	"""

	def __init__(self, units, 
						activation=None, 
						weights_initialiser="GlorotUniform", 
						bias_initialiser="GlorotUniform", 
						use_bias=True, 
						name=None, **kwargs) -> None:
		super().__init__(name, **kwargs)

		self._units = units
		self.activation = set_activation(activation, self.custom_objs)

		self.weights_initialiser = set_initialiser(weights_initialiser)
		self.bias_initialiser = set_initialiser(bias_initialiser)

		self.use_bias = use_bias

	@property
	def units(self): return self._units
	@units.setter
	def units(self, value): ...

	def build(self, input_shape, input_dtype, **kwargs):
		self.weights = self.add_weight(
			self.weights_initialiser(shape=(self.units, input_shape[0]))(), dtype=input_dtype) 

		self.bias = self.add_weight(
			self.bias_initialiser(shape=(self.units, 1))() if self.use_bias else 0, dtype=input_dtype)

		return super().build(input_shape, input_dtype, **kwargs)

	def call(self, x):
		output = (self.weights @ x) + self.bias
		
		return self.activation(output)
	
	def config(self):
		self.set_config(rename_key(super().config(), "_units", "units"))
		self.__dict__["bias_initialiser"] = self.bias_initialiser.__name__
		self.__dict__["weights_initialiser"] = self.weights_initialiser.__name__
		self.__dict__["activation"] = self.activation.__class__.__name__
		return self.__dict__

# aliases
FC = FullyConnected
Dense = FullyConnected


import numpy as np
from bren.nn.layers.Layer import Layer
from bren.autodiff.operations.ops import custom_gradient


def elu_grad(a, alpha, dout, value):
	return [np.multiply(dout, np.where(a > 0, 1, alpha * np.exp(a)))]
@custom_gradient(elu_grad)
def elu(a, alpha): 
	return np.where(a > 0, a, alpha * (np.exp(a) - 1))


class ELU(Layer):
	"""
	Perform the ELU activation function on the input. 
	For input values greater than 1 the output will be linear (`x`), while negative values will be computed as `alpha * np.exp(x) - 1` with the alpha hyperparameter determining what negative value the function approches for more and more negative values of the inputs.

	Paratmeters
	-----------
	alpha (`float`): Hyperparameter which determines the negative value which the funcion approached for more and more negative value of x
	x (`br.Variable`): The input Array.
	name (`str`): The name of the activation.
	"""

	def __init__(self, alpha=1, name=None, **kwargs) -> None:
		self.alpha = alpha
		super().__init__(name, **kwargs)

	def call(self, x): return elu(x, self.alpha)


import numpy as np
from bren.nn.initialisers.Initialiser import Initialiser


class GlorotUniform(Initialiser):
	"""
	Returns values taken from a uniform distribution with endpoints `r` and `-r` where `r = (6 / fan_in + fan_out)` with `fan_in` being the number of inputs while `fan_out` being the number of outputs
	"""

	def __init__(self, shape, dtype="float32", **kwargs) -> None:
		super().__init__(shape, dtype, **kwargs)

		self.fan_in, self.fan_out = kwargs.get("inout") or shape
		fan_avg = (self.fan_in + self.fan_out) / 2
		self.r = np.sqrt(3 / fan_avg)

	def call(self, *args, **kwargs):
		return np.random.uniform(-self.r, self.r, size=self.shape)


import bren.autodiff.nodes as nodes


class Var(nodes.Node):
	count = 0

	def __init__(self, value, name=None) -> None:
		self.name = name or self.set_name(Var) 
		Var.count += 1
		if nodes.Graph._g: nodes.Graph._g.add(self)
		super().__init__(value, self.name)
	
	def __repr__(self) -> str:
		return f"<{Var.__name__} name={self.name} value={self.value}>"


import numpy as np
import bren as br
from bren.nn.metrics import get_metric
from bren.nn.losses import get_loss
from bren.nn.optimisers import get_optimiser
import pickle
import copy


def set_metric(metric, custom_obs={}):
	out = None
	if type(metric) is str:
		try:
			out = get_metric(metric)()
		except KeyError:
			try:
				out = set_metric(custom_obs[metric])
			except KeyError:
				raise KeyError(f"custom object {metric} could not be found")
	elif type(metric).__name__ == "function":
		try: 
			out = get_metric(metric.__name__)()
		except KeyError:
			out = br.nn.metrics.metric_from_loss(metric)()
	elif issubclass(type(metric), br.nn.metrics.Metric):
		out = metric
	elif isinstance(metric, type):
		out = metric()
	else: out = get_metric(metric.__name__)()

	return out

def set_loss(loss, custom_obs={}):
	out = None

	if type(loss) is str:
		try: 
			out = get_loss(loss)()
		except KeyError:
			try: 
				out = set_loss(custom_obs[loss])
			except KeyError:
				raise KeyError(f"custom object {loss} could not be found")
	elif type(loss).__name__ == "function":
		try:
			out = get_loss(loss.__name__)()
		except KeyError: 
			out = br.nn.losses.Loss(name=loss.__name__, func=loss)
	elif issubclass(type(loss), br.nn.losses.Loss):
		out = loss
	elif isinstance(loss, type):
		out = loss()
	else: out = get_loss(loss.__name__)()

	return out

def set_optimiser(optimiser):
	out = None

	if type(optimiser) is str:
		out = get_optimiser(optimiser)()
	elif issubclass(type(optimiser), br.nn.optimisers.Optimiser):
		out = optimiser
	else: out = get_optimiser(optimiser.__name__)()

	return out

class Model(object):
	"""
	The base `Model` class
	"""

	def __init__(self, **kwargs) -> None:
		self.training = False
		self.assembled = False
		self.built = False

		self.trainable = kwargs.get("trainable") or []
		self.params = []
		self.serialised = False
		self.copy = None 

		self.custom_obs = None

	@property
	def config(self): return self.__config
	@config.setter
	def config(self, val): ...

	def add_config(self, key, value): 
		"""
		Adds a key value pair to the config

		Parameters
		----------
		key: The key
		value: The value
		"""
		self.__config.update({**self.config, key: value})

	def assemble(self, loss=None, optimiser=None, metrics=[], **kwargs):
		"""
		Establishes the loss function, the optimiser and the metrics of the model (the loss function is by default added to the metrics).

		Parameters
		----------
		loss (`str`, `function`, `br.nn.losses.Loss`): The loss function of the model
		optimiser (`str`, `br.nn.optimisers.Optimiser`): The optimiser of the model
		metrics (`list`): A list of all of the metrics which you want to be displayed
		"""

		self.optimiser = None
		self.metrics = []
		self.loss = None

		self.assembled = True
		self.loss = set_loss(loss, self.custom_obs)
		self.optimiser = set_optimiser(optimiser)
		self.metrics.append(set_metric(self.loss.func, custom_obs=self.custom_obs))

		for metric in metrics:
			self.metrics.append(set_metric(metric, custom_obs=self.custom_obs))


	def set_custom_object(self, obs): 
		"""
		Updates the value of the of the custom objects attirbute, which is initiallly set to `None`

		Parameters
		----------
		obs (`dict`): The key value pairs of the custom components
		"""
		self.custom_obs = obs

	def call(self, x, training=None):
		"""
		Where the functionality of the model takes place.

		Parameters
		----------
		x (`br.Variable`): the inputs 
		training (`bool`): A boolean value which is passed to the layers which specifies if the model is training or not
		"""
		...

	def build(self, input):
		"""
		Builds the model. Performs one forward pass with the first value in the training data.

		Parameters
		----------
		input (`br.Variable`): the inputs to the model,
		"""
		self.built = True
		self.call(input[0]) 

	def fit(self, x, y, epochs=1, shuffle=False, batch_size=1, print_details=True, *args, **kwargs):
		"""
		Trains the model. 

		Parameters
		----------
		x (`br.Variable`): The features of the training data.
		y (`br.Variable`): The lables of the training data.
		epochs (`int`): The number of iterations of the training data which.
		shuffle (`bool`): Whether the trainng data should be shuffled before being passed through the model.
		print_details (`bool`): Whether the loss or metrics values are printed out while training.
		"""

		if not self.assembled: raise RuntimeError("The model should be assembled before you can train it.")
		if not self.built: 
			self.build(x)
		
		for params in self.trainable: self.params.extend(params.values())

		X_batch = br.nn.preprocessing.split_uneven(x, batch_size)[..., np.newaxis]
		Y_batch = br.nn.preprocessing.split_uneven(y, batch_size)[..., np.newaxis]

		if shuffle: br.nn.preprocessing.shuffle(Y_batch)

		for i in range(1, epochs + 1):
			self.__train_batch(X_batch, Y_batch)

			if print_details:
				print(f"EPOCH {i}/{epochs}: ", end="")

				for metric in self.metrics:
					print(metric.__class__.__name__, ":", metric.result().numpy(), end=" - ")
					metric.reset()
				print()
		
	def add_weight(self, val, **kwargs):
		self.trainable.append(val)

	def __forward_update(self, X, Y, training=None):
		loss = []
		for x, y in zip(X, Y):
			Z = self.call(x, training=training)
			loss.append(self.loss(Z, y))
			
			for metric in self.metrics:
				metric.update(Z, y)

		return np.sum(loss)

	def __train_batch(self, X_batches, Y_batches):
		for x, y in zip(X_batches, Y_batches):
			with br.Graph() as g:
				loss = self.__forward_update(x, y, training=True)
				
				grad = g.grad(loss, self.params)
				self.optimiser.apply_gradients(self.params, grad)
		
	def predict(self, X): 
		"""
		Performs a forward pass on the given data.

		Parameters 
		----------
		X (`br.Variable`): The Testing data.
		"""

		output = []
		for x in X:
			output.append(self.call(x[..., np.newaxis], training=False).numpy())

		return np.array(output)
	
	def serialise(self):
		"""
		Serialises the model's attributes
		"""

		self.__config = {
			"optimiser": self.optimiser.__class__.__name__,
			"loss": self.loss.name
			}
		
		self.__config["metrics"] = []
		for metric in self.metrics:
			self.__config["metrics"].append(metric.__class__.__name__)

		try: self.__config["metrics"].remove(self.loss.name)
		except ValueError: pass

		for key in self.config.keys():
			try: del self.__dict__[key]
			except KeyError: ...


	def save(self, filepath): 
		"""
		Saves the model as a pickle file to the specified file path.

		Parameters
		----------
		filepath (`str`): The file path to the model
		"""

		cop = copy.deepcopy(self)
		cop.serialise()

		with open(filepath, "wb") as f:
			pickle.dump(cop, f)



from bren.autodiff.nodes.Node import Node
from bren.autodiff.nodes.Graph import Graph
from bren.autodiff.nodes.Operator import Operator
from bren.autodiff.nodes.Var import Var
from bren.autodiff.nodes.Const import Const


__all__ = [Node, Graph, Operator, Var, Const]



from bren import Variable


class Layer(object):
	"""
	The base `Layer` class.

	Parameters
	----------
	name (`str`): The name of the layer.
	"""

	count = 0

	def __init__(self, name=None, **kwargs) -> None:
		self.custom_objs = kwargs.get("custom_obs")
		self.trainable = kwargs.get("params", {})
		self.name = name or self.set_name(Layer)
		self.built = False
		Layer.count += 1

	def __call__(self, inp, training=None, **kwargs): 
		# print(self.name, self.__class__, self.built)
		if not self.built:
			# print("This was called")
			self.build(input_shape=inp.shape, input_dtype=inp.dtype)
		return self.call(inp)
		 
	def __delattr__(self, __name) -> None: ...
	
	def build(self, input_shape, input_dtype, **kwargs):
		"""
		Builds the model

		Parameters
		----------
		input_shape (`tuple`): the shape of the input value
		input_dtype (`str`): the data type of the input value
		"""
		for key, value in zip(self.__dict__.keys(), self.__dict__.values()):
			if type(value) is Variable: self.trainable[key] = value
		
		self.built = True

	def set_built(self, bool): self.built = bool

	def call(self, x): 
		"""
		Where the computation of the layer will take place

		Parameters
		----------
		x (`br.Variable`): the input to the layer
		"""
		...

	def add_weight(self, val, **kwargs):
		"""
		Creates and returns a Variable

		Parameters
		----------
		val (`list`, `np.ndarray`, `br.Variable`): the value of the array
		"""
		var = Variable(val, **kwargs)
		return var

	def set_weights(self, params):
		self.trainable = params

	def get_weights(self):
		return self.trainable

	def set_name(self, cls): return f"{cls.__name__}/{cls.count}"

	def set_config(self, dictionary): self.__dict__ = dictionary 

	def config(self):
		"""
		Returns the model config in the form of a dictionary
		"""
		if "layer" in self.__dict__.keys():
			raise AttributeError(f"The attribute 'layer' is reserved for the model's internals, please change the attribute name")

		return {"layer": self.__class__.__name__, **self.__dict__}
	




import numpy as np
import requests, gzip, os, pathlib
from bren import Variable


path = os.path.join(pathlib.Path(__file__).parent.resolve(), "data")


def fetch(url):
	"""
	Fetches files from a given url
	"""
	if os.path.isfile(path):
		with open(path, "rb") as f:
			data = f.read()
	else:
		with open(path, "wb") as f:
			data = requests.get(url).content
			f.write(data)
	os.remove(path)
	return np.frombuffer(gzip.decompress(data), dtype=np.uint8).copy()


def load_mnist(dtype="float64"):
	"""
	Loads the mnist dataset and converts the data into a tuple of `br.Variable`

	Parameters
	----------
	dtype (`str`): Determines the data type of the dataset

	Returns
	-------
	Returns a tuple of length 4 with the training features, training lables, testing features and testing lables respectively.
	"""
	X = fetch("http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz")[0x10:].reshape((-1, 28, 28)) / 255.
	Y = fetch("http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz")[8:] 
	X_test = fetch("http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz")[0x10:].reshape((-1, 28, 28)) / 255.
	Y_test = fetch("http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz")[8:]
	return (Variable(X, dtype=dtype), Variable(Y, dtype=dtype), 
		 Variable(X_test, dtype=dtype), Variable(Y_test, dtype=dtype))



import bren.autodiff.nodes as nodes


def return_values(x): return x.value

class Operator(nodes.Node):

	def __init__(self, inputs, value=None, name="Operator", **kwargs) -> None:
		self.inputs = inputs
		self.scalars = list(map(return_values, self.inputs))
		self.grad = kwargs.get("grad")
		super().__init__(value, name)

	def __call__(self, dout): 
		self.gradient = 0
		return self.grad(*self.scalars, dout=dout, value=self.value)

	def __repr__(self) -> str:
		return f"<{self.name.capitalize()} inputs={self.inputs}>"


class Initialiser(object):
	"""
	The base Initialiser class
	"""
	def __init__(self, shape, dtype="float32", **kwargs) -> None:
		self.shape = shape
		self.dtype = dtype

	def __call__(self, *args, **kwargs): 
		return self.call(**self.__dict__).astype(self.dtype)

	def call(self, shape, *args, **kwargs):
		"""
		Where the functionaliry of the initialiser takes place

		Parameters
		----------
		shape (`tuple`): the shape of the input
		"""
		...


def initialiser_from_func(func):
	"""
	Used to produce an Initialiser class with a custom function

	Parameters
	----------
	func (`function`): the intitalisation function
	"""
	class Initialiser(object):
		def __init__(self, shape, dtype="float32", **kwargs) -> None:
			self.shape = shape
			self.dtype = dtype

		def __call__(self, *args, **kwargs): 
			return func(**self.__dict__).astype(self.dtype)

	return Initialiser


from bren.nn.models import Model
import numpy as np
import bren as br


class Sequential(Model):
    """
    The `Sequential` model takes in a list of layers and passes the data through each of the layers consecutively.

    Parameters 
    ----------
    layers (`list`): The list of layers.
    """

    def __init__(self, layers, **kwargs) -> None:
        super().__init__(**kwargs)
        self.layers = layers

    def build(self, input):
        super().build(input)
        
        for layer in self.layers:
            self.add_weight(layer.trainable)

    def call(self, x, training=None):
        Z = x
        for layer in self.layers:
            Z = layer(Z, training=training)
        return Z
    
    def serialise(self):
        for i, layer in enumerate(self.layers):
            self.layers[i] = layer.config()
        return super().serialise()
    
    def save(self, filepath):
        return super().save(filepath)


from bren.nn.models.Model import Model
from bren.nn.models.Sequential import Sequential
from bren.nn.utils import AliasDict
from bren.nn.layers import __all__ as layers
import pickle

__all__ = [Model, Sequential]

MODELS = AliasDict({})


def get_model(name): return MODELS[name]


def deserialise_sequential_layers(model, custom_objects):
    OBJECTS = {**custom_objects}
    for layer in layers: OBJECTS[layer.__name__] = layer

    model_layers = model.layers
    
    for i, layer in enumerate(model.layers):
        L = OBJECTS[layer["layer"]](**layer, custom_obs=custom_objects, params=layer.get("trainable", []))
        model_layers[i] = L
        model_layers[i].set_built(True)
        try: model_layers[i].activation.set_built(True)
        except AttributeError: pass 
        model_layers[i].__dict__ = {**model_layers[i].__dict__, **layer["trainable"]} 


def load_model(filepath, custom_objects={}):
    with open(filepath, "rb") as f:
        model = pickle.load(f)

    model.set_custom_object(custom_objects)

    if model.__class__.__name__ == "Sequential":
        deserialise_sequential_layers(model, custom_objects)

    model.assemble(**model.config)
    
    return model


for cls in __all__:
    MODELS[cls.__name__] = cls


from bren.nn.metrics import Metric
import numpy as np

class Loss(Metric):
	"""
	The base `Loss` class

	Parameters
	----------
	name (`str`): The name of the function.
	"""

	def __init__(self, name=None, **kwargs) -> None: 
		self.func = kwargs.get("func", None)
		self.name = name or self.__class__.__name__

	def __call__(self, y_pred, y_true, *args, **kwargs): 
		out = self.func(y_pred, y_true, *args, **kwargs)
		return np.sum(out)
		


from bren.nn.optimisers.Optimiser import Optimiser
from bren.nn.optimisers.SGD import SGD
from bren.nn.optimisers.AdaGrad import AdaGrad
from bren.nn.optimisers.RMSProp import RMSProp
from bren.nn.optimisers.Adam import Adam
from bren.nn.utils import AliasDict


__all__ = [Optimiser, SGD, AdaGrad, RMSProp, Adam]


OPTIMISERS = AliasDict()

for cls in __all__:
    OPTIMISERS[cls.__name__] = cls

OPTIMISERS.add("AdaGrad", "adagrad")
OPTIMISERS.add("Adam", "adam")
OPTIMISERS.add("SGD", "sgd", "StochasticGradientDescent", "stochastic_gradient_descent")
OPTIMISERS.add("RMSProp", "rmsprop", "RootMeanSquaredPropogation", "root_mean_squared_propogation")

def get_optimiser(name):
    return OPTIMISERS[name]


from bren.nn.metrics import Metric
import numpy as np


class Accuracy(Metric):
	"""
	`Accuracy` calculates how amount of times which `y_pred` is equal to `y_true` with a given `leniency` and adds it to the variable `total`. When the result function is called, `total/count` is returned.
	
	Parameters
	----------
	precision (`int`): the decimal place at which `y_pred` is rounded before begin compared to `y_true`. They higher the precision, the more accurate `y_pred` will need to be to `y_true` to get a high accuracy.
	"""

	def __init__(self, precision=2) -> None:
		super().__init__()
		self.total = self.add_weight(0)
		self.count = self.add_weight(0)
		self.precision = precision

	def update(self, y_pred, y_true, weights=None, **kwargs):
		acc = np.sum(y_true == np.round(y_pred, decimals=self.precision)) / len(y_true)
		self.total.assign_add(acc)
		self.count.assign_add(1)

	def result(self):
		return self.total / self.count


from bren.nn.optimisers import Optimiser
from bren import Constant
import numpy as np


class RMSProp(Optimiser):
	def __init__(self, learning_rate=0.001, beta=0.9, epsilon=10e-10) -> None:
		super().__init__()
		self.learning_rate = learning_rate
		self.beta = beta
		self.epsilon = epsilon
		self.s = []

	def initialise(self, len):
		for _ in range(len): self.s.append(Constant(0))

	def apply_gradients(self, vars, grads, **kwargs):
		if len(self.s) == 0: self.initialise(len(vars))
		self.update(vars, grads, self.s)

	def update(self, var, grad, s):
		_s = s.assign((self.beta * s) + (1 - self.beta) * (grad ** 2))
		_t = (self.learning_rate * grad) / (np.sqrt(_s + self.epsilon))
		var.assign_sub(_t / (1 - self.beta))


import bren.nn.layers as layers
import bren.nn.initialisers as initialisers
import bren.nn.activations as activations
import bren.nn.optimisers as optimisers
import bren.nn.losses as losses
import bren.nn.metrics as metrics
import bren.nn.models as models
import bren.nn.preprocessing as preprocessing
import bren.nn.datasets as datasets

__all__ = [layers, initialisers, activations, optimisers, losses, metrics, models, preprocessing, datasets]



import numpy as np


class Optimiser(object):
	"""
	The base `Optimiser` class.
	"""

	def __init__(self) -> None: 
		self.update = np.vectorize(self.update, cache=True, otypes=[list])

	def apply_gradients(self, vars, grads, **kwargs): 
		"""
		Updates the given variables with the given gradients of the variables with respect to the loss, calculated by autodiff.

		Parameter
		---------
		vars (`list`): A list of the trainable Variables.
		grads (`list`): A list of the respective Gradients.

		"""
		...

	def update(*args, **kwargs): 
		"""
		A vectorised function which performs the updating of the value of the variables (called within the `apply_gradients` function).
		"""
		...


import numpy as np
from bren.autodiff.nodes import Var, Const
from bren.core.Array import Array


class Variable(Array):
	"""
	Produces a `Variable` object (wrapped numpy array) which keeps track of functions which were performed on it.

	Parameters: 
	-----------
	value (`np.ndarray`, `list`): Value to be stored by the object
	dtype (`str`): The data type of the array
	name (`str`): The name of the array
	"""
	def __init__(self, value, dtype="float32", name=None, **kwargs) -> None:
		super().__init__(value, dtype, name, **kwargs)
		self._source = kwargs.get("source") or Var(self._i, name=self.name)


class Constant(Array):
	"""
	Produces a `Constant` object (wrapped numpy array) which does not have a gradient.

	Parameters: 
	-----------
	value (`np.ndarray`, `list`): Value to be stored by the object
	dtype (`str`): The data type of the array
	name (`str`): The name of the array
	"""

	def __init__(self, value, dtype="float32", name=None, **kwargs) -> None:
		super().__init__(value, dtype, name, **kwargs)
		self._source = kwargs.get("source") or Const(self._i, name=self.name)



from bren.nn.activations.Tanh import Tanh, tanh
from bren.nn.activations.ReLU import ReLU
from bren.nn.activations.ELU import ELU
from bren.nn.activations.Linear import Linear
from bren.nn.utils import AliasDict
from bren.nn.activations.Activation import Activation
from bren.nn.layers.Softmax import Softmax


__all__ = [Tanh, ReLU, ELU, Activation, Softmax]


ACTIVATIONS = AliasDict({ None: Linear })

for cls in __all__:
	ACTIVATIONS[cls.__name__] = cls

ACTIVATIONS.add(ReLU.__name__, "relu")
ACTIVATIONS.add(ELU.__name__, "elu")
ACTIVATIONS.add(Tanh.__name__, "tanh")

def get_activation(name):
	return ACTIVATIONS[name]


class AliasDict(dict):
	def __init__(self, *args, **kwargs):
		dict.__init__(self, *args, **kwargs)
		self.aliases = {}

	def __getitem__(self, __key):
		return dict.__getitem__(self, self.aliases.get(__key, __key))

	def __setitem__(self, __key, __value) -> None:
		return dict.__setitem__(self, self.aliases.get(__key, __key), __value)

	def add(self, __key, *aliases):
		for alias in aliases:
			self.aliases[alias] = __key


def progress_bar(metrics):
	...


def rename_key(dictionary, k_old, k_new):
	"""
	Renames the key in a dictionary without editting its position.

	Parameters
	----------
	dictionary (`dict`): The dictionary to be editted.
	k_old (`str`): The old key name.
	k_new (`str`): The new key name.

	Returns
	-------
	Returns a new dictionary with the renamed key name
	"""

	pos = list(dictionary.keys()).index(k_old)
	value = (k_new, dictionary[k_old])
	dictionary.pop(k_old)

	items = list(dictionary.items())
	items.insert(pos, value)

	return dict(items)




from bren.nn.metrics.Metric import Metric
from bren.nn.metrics.Accuracy import Accuracy
from bren.nn.metrics.MeanSquaredError import MeanSquaredError
from bren.nn.metrics.CategoricalCrossEntropy import CategoricalCrossEntropy
from bren.nn.utils import AliasDict
import numpy as np


__all__ = [Metric, Accuracy, MeanSquaredError, CategoricalCrossEntropy]


METRICS = AliasDict()

for cls in __all__:
    METRICS[cls.__name__] = cls
    
METRICS.add(Accuracy.__name__, "accuracy")
METRICS.add(MeanSquaredError.__name__, "mse", "mean_squared_error", "MSE")
METRICS.add(CategoricalCrossEntropy.__name__, "categorical_cross_entropy")


def get_metric(name):
	return METRICS[name]


def metric_from_loss(func, name=None):
	class MetricFromLoss(Metric):
		def __init__(self) -> None:
			super().__init__()
			self.__class__.__name__ = name or func.__name__
			self.total = self.add_weight(0)
			self.count = self.add_weight(0)

		def update(self, y_pred, y_true, weights=None):
			self.total.assign_add(func(y_pred, y_true))
			self.count.assign_add(1)
		
		def result(self):
			return np.sum(self.total / self.count)

	return MetricFromLoss
		


from bren.nn.layers import Layer


class Activation(Layer):
	"""
	The `Activation` class is a child of the `Layer` class and is mainly used to produce custom activation funtions.

	Parameters
	__________
	func (`function`): the activation function
	name (`str`): the name of the `Activation` object
	"""

	def __init__(self, func=None, name=None, **kwargs) -> None:
		self.func = func
		super().__init__(name, **kwargs)
		self.set_built(True)
		self.__class__.__name__ = func.__name__

	def call(self, inp, training=None, **kwargs):
		return self.func(inp)



from bren.nn.optimisers import Optimiser
from bren import Constant


class SGD(Optimiser):
	def __init__(self, learning_rate=0.01, beta=0) -> None:
		super().__init__()
		self.learning_rate = learning_rate
		self.beta = beta 
		self.momentum = []

	def initialise(self, len):
		for _ in range(len): self.momentum.append(Constant(0))

	def apply_gradients(self, vars, grads, **kwargs): 
		if len(self.momentum) == 0: self.initialise(len(vars))
		self.update(vars, grads, self.momentum)

	def update(self, var, grad, momentum): 
		m = momentum.assign((self.beta * momentum) - (self.learning_rate * grad)) 
		var.assign_add(m)




import bren.autodiff.nodes as nodes


class Const(nodes.Node):
	count = 0

	def __init__(self, value, name=None) -> None:
		self.name = name or self.set_name(Const) 
		Const.count += 1
		if nodes.Graph._g: nodes.Graph._g.add(self)
		super().__init__(value, self.name)
	
	def __repr__(self) -> str:
		return f"<{Const.__name__} name={self.name} value={self.value}>"

	@nodes.Node.gradient.setter
	def gradient(self, value): ...


import numpy as np
from bren.autodiff.nodes import Graph
from bren.autodiff.operations.ops import OPS
import typing


def make_ops_source(name, **kwargs):
	try: 
		if Graph._g: return OPS.get(name, OPS["nongrad"])(**kwargs)
		else: return None
	except KeyError:
		raise RuntimeError(f"no gradient found for operation {name}")


class Array(np.lib.mixins.NDArrayOperatorsMixin, typing.Sequence):	
	def __init__(self, value, dtype="float32", name=None, **kwargs) -> None:
		self.dtype = dtype
		if issubclass(type(value), Array):
			self._i = np.array(value._i, dtype=self.dtype)
		else: self._i = np.array(value, dtype=self.dtype)
		self.shape = self._i.shape
		self.ndim = self._i.ndim
		self.size = self._i.size
		self.name = name
		self.frozen = False
		super().__init__()

	@property
	def length(self): return self.shape[0]
	@length.setter
	def length(self): ...

	def freeze(self): self.frozen = True
	def unfreeze(self): self.frozen = False

	def __len__(self): 
		if not self.frozen: return self.shape[0]

	def __repr__(self) -> str:
		return f"<{self.__class__.__name__} value={self._i} dtype={self.dtype}>"
	
	def __getitem__(self, key):
		return self.__class__(self._i[key], dtype=self.dtype)
	def __setitem__(self, key, value):
		self._i[key] = value

	def __iter__(self):
		self._idx = 0
		return self
	
	def __next__(self):
		if self._idx < len(self._i):
			x = self._i[self._idx]
			self._idx += 1
			return x
		raise StopIteration

	def __array_function__(self, func, types, args, kwargs):
		scalars, sources = self.__set_source_scalar(args)

		val = func(*scalars, **kwargs)
		source = make_ops_source(func.__name__, inputs=sources, value=val)

		return self.__class__(val, dtype=self.dtype, source=source)

	def __array_ufunc__(self, ufunc, method, *inps, **kwargs):
		if method == "__call__":
			scalars, sources = self.__set_source_scalar(inps)

		val = ufunc(*scalars, **kwargs)
		source = make_ops_source(ufunc.__name__, inputs=sources, value=val)
		return self.__class__(val, dtype=self.dtype, source=source)	

	def __set_source_scalar(self, inps):
		scalars = []
		sources = []

		for inp in inps:
			val = inp
			
			if not issubclass(type(inp), self.__class__):
				try: val = self.__class__(inp)
				except:
					raise ValueError(f"Cannot convert {inp} into type {self.__class__.__name__}")

			scalars.append(val._i)
			sources.append(val._source)

		return scalars, sources

	def append(self, values, **kwargs):
		"""
		Inserts values at the end of the array.
		
		Parameters
		----------
		value: value to be added
		"""
		self._i = np.append(self._i, values, *kwargs)

	def assign(self, value):
		"""
		Reassigns the array's value. Does not return a clone of the array.

		Parameters
		----------
		value: value to be assigned to the array

		Returns
		-------
		Returns the numpy array which is held in the array
		"""
		if issubclass(type(value), Array): value = value._i
		self._i = value
		self._source.value = self._i

		return self._i

	def assign_add(self, value): 
		"""
		Adds and assigns the result to the array.

		Parameters
		----------
		value: value to be added

		Returns
		-------
		Returns the numpy array which is held in the array
		"""
		return self.assign(self._i + value)
	
	def assign_sub(self, value): 
		"""
		Subtracts and assigns the result to the array.

		Parameters
		----------
		value: value to be subtracted

		Returns
		-------
		Returns the numpy array which is held in the array
		"""
		return self.assign(self._i - value)

	def clone(self): 
		"""
		Returns a clone of the array.
		
		[NOTE] any built up computation graphs from the origional array will be discarded.
		"""
		return self.__class__(self._i, self.dtype, self.name)

	def numpy(self, dtype=None): 
		"""
		Returns a numpy array of the value
		"""
		return self._i.astype(dtype or self.dtype)

	def flatten(self): 
		"""
		Returns a 1D version of the array
		"""
		return self._i.flatten()

	def astype(self, dtype): 
		"""
		Changes the data type of the array and returns the array itsef.
		"""
		self.dtype = dtype
		return self

	@property
	def T(self): 
		"""
		Returns the transposed version of the array.
		"""
		return np.transpose(self)
	@T.setter
	def T(self): ...




from bren.nn.losses import Loss
import numpy as np


def categorical_cross_entropy(y_pred, y_true, epsilon=1e-8): 
	return -(np.sum(y_true * np.log(y_pred + epsilon)))

class CategoricalCrossEntropy(Loss):
	"""
	`CategoricalCrossEntropy` computes the loss for multiclass classification, with `loss=-(sum(y_true * log(y_pred + epsilon)))` 
	"""

	def __init__(self, name=None, **kwargs) -> None:
		super().__init__(name, **kwargs)
		self.func = categorical_cross_entropy



from setuptools import setup, find_packages

from codecs import open
from os import path

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'PACKAGE.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="bren",
    version="0.1.2",
    description="A simple numpy based neural network library inspired by Tensorflow/Keras.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Om Panchal",
    author_email="om.panchal2022@gmail.com",
    maintainer="Om Panchal",
    maintainer_email="om.panchal2022@gmail.com",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
	],
    install_requires=["numpy", "requests"]
)


import numpy as np
from bren.nn.layers.Layer import Layer


def tanh(x): return np.tanh(x)

class Tanh(Layer):

	"""
	Performs `np.tanh(x)` on the inputs

	Parameters
	----------
	x (`br.Variable`): The input Array
	name (`str`): The name of the activation.	
	"""

	def __init__(self, name=None, **kwargs) -> None:
		super().__init__(name, **kwargs)

	def call(self, x): return tanh(x)


from bren.nn.losses.Loss import Loss
from bren.nn.losses.MeanSquaredError import mse, mean_squared_error, MeanSquaredError, MSE
from bren.nn.losses.CatrgoricalCrossEntropy import categorical_cross_entropy, CategoricalCrossEntropy
from bren.nn.utils import AliasDict


__all__ = [Loss, MeanSquaredError, CategoricalCrossEntropy]

LOSSES = AliasDict()

for cls in __all__:
    LOSSES[cls.__name__] = cls

LOSSES.add(MeanSquaredError.__name__, "MSE", "mse", "mean_squared_error")
    
def get_loss(name):
	return LOSSES[name]


from bren.nn.metrics import Metric
import numpy as np


class CategoricalCrossEntropy(Metric):
    """
    The metric equivalient of the `CategoricalCrossEntropy` loss function which computes the loss for multiclass classification, with `loss=-(sum(y_true * log(y_pred)))`
    """

    def __init__(self) -> None:
        super().__init__()
        self.total = self.add_weight(0)
        self.count = self.add_weight(0)
    
    def update(self, y_pred, y_true, weights=None, **kwargs):
        self.total.assign_add(kwargs.get("loss", -(np.sum(y_true * np.log(y_pred + 10e-10)))))
        self.count.assign_add(1)

    def result(self):
        return self.total / self.count


import random
import bren as br
import numpy as np

def shuffle(*arrays, seed_range=1000):
    """
    Randomly shuffles the values in the arrays given, using the same seed throughout.

    Parameters
    ----------
    seed_range (`int`): the seed is a positive integer, and is chosen randomly between the range of 0 and `seed_range` 
    """
    seed = random.randint(0, seed_range)

    for array in arrays:
        np.random.seed(seed)
        np.random.shuffle(array)

    
def split_uneven(array, split):
    """
    Splits an array into a specified number of sub-arrays and appends the remainders to the end of the array.

    Parameters
    ----------
    array (`br.Variable`, `list`): The array.
    split (`int`): The number of sub arrays which you want to split the array into.

    Returns
    -------
    Returns a `br.Variable` of the array split into the number of sub arrays as by `split` along with the remainders at the end of the array
     """

    remainder = len(array) % split
    return br.Variable(np.split(array[:len(array) - remainder], split))


import numpy as np
from bren.nn.layers.Layer import Layer
from bren.autodiff.operations.ops import custom_gradient


def relu_grad(a, leak, dout, value): 
	return [np.multiply(dout, a > 0)]
@custom_gradient(relu_grad)
def relu(a, leak): return np.maximum(a * leak, a)


class ReLU(Layer):
	"""
	[NOTE]: This is also the leaky ReLU function, change the leak parameter to change to leaky ReLU. By default the `leak` parameter for this function is 0 to simulate the nature of a regular ReLU function.

	Performs the ReLU activation function on the inputs.
	For positive values the output is linear (`x`) while for negative values the output is `x * leak` where leak is the gradient of the straight line for values smaller than 0. 

	Parameters
	----------
	leak (`float`): the gradient of the straight lines for values smaller than 0
	x (`br.Variable`): The input Array
	name (`str`): The name of the activation.
	"""

	def __init__(self, leak=0, name=None, **kwargs) -> None:
		self.leak = leak
		super().__init__(name, **kwargs)

	def call(self, x): return relu(x, self.leak)


from bren.nn.losses import Loss
import numpy as np


def mse(y_pred, y_true): return np.mean((y_pred - y_true) ** 2)

class MeanSquaredError(Loss):
	"""
	`MeanSquaredError` computes the loss as `loss=mean((y_pred - y_true) ** 2)`
	"""
	
	def __init__(self, name=None, **kwargs) -> None:
		super().__init__(name, **kwargs)
		self.func = mse


MSE = MeanSquaredError
mean_squared_error = mse


from bren.nn.metrics import Metric
import numpy as np


class MeanSquaredError(Metric):
	"""
	The metric equivalent of the `MeanSquaredError` loss function which computes the loss as `loss=mean((y_pred - y_true) ** 2)`
	"""

	def __init__(self) -> None:
		super().__init__()
		self.total = self.add_weight(0)
		self.count = self.add_weight(0)
		
	def update(self, y_pred, y_true, weights=None, **kwargs): 
		self.total.assign_add(kwargs.get("loss", np.sum((y_true - y_pred) ** 2))) 
		self.count.assign_add(1)

	def result(self): return self.total / self.count



from bren.nn.layers.Layer import Layer
from bren.nn.layers.FullyConnected import Dense, FullyConnected, FC
from bren.nn.layers.Flatten import Flatten
from bren.nn.layers.Softmax import Softmax

__all__ = [Layer, Dense, FullyConnected, FC, Flatten, Softmax]




from bren.nn.initialisers.Initialiser import Initialiser, initialiser_from_func
from bren.nn.initialisers.GlorotNormal import GlorotNormal
from bren.nn.initialisers.GlorotUniform import GlorotUniform
from bren.nn.initialisers.HeNormal import HeNormal
from bren.nn.initialisers.HeUniform import HeUniform
from bren.nn.utils import AliasDict


__all__ = [Initialiser, GlorotNormal, GlorotUniform, HeNormal, HeUniform, initialiser_from_func]


INITIALISERS = AliasDict({})

for cls in __all__:
	INITIALISERS[cls.__name__] = cls

INITIALISERS.add(GlorotNormal.__name__, "glorot_normal", "glorotnormal")
INITIALISERS.add(GlorotUniform.__name__, "glorot_uniform", "glorotuniform")
INITIALISERS.add(HeNormal.__name__, "he_normal", "henormal")
INITIALISERS.add(HeUniform.__name__, "he_uniform", "heuniform")

def get_initialiser(name): return INITIALISERS.get(name, name)